library simple_billing;

import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:flutter/material.dart';
import 'package:flutter_inapp_purchase/flutter_inapp_purchase.dart';
import 'package:googleapis_auth/auth.dart';
import 'package:googleapis_auth/auth_io.dart';
import 'package:hive/hive.dart';
import 'package:hive_flutter/hive_flutter.dart';
import 'package:http/http.dart' as http;
import 'package:path_provider/path_provider.dart' as path_provider;

part 'local_data.dart';
part 'pop_up_widget.dart';

class SimpleBilling {
  //! Public Vars
  static Set<PurchasedItem> purchases = {};

  //! Private Vars
  static StreamSubscription _purchaseUpdatedSubscription;
  static StreamSubscription<String> _purchasePromotedSubscription;
  static StreamSubscription _purchaseErrorSubscription;
  static void Function(PurchasedItem productId) _onPurchaseUpdated;
  static List<IAPItem> _productsItems = [];
  static List<IAPItem> _subscriptionsItems = [];
  static BuildContext _context;
  static final Set<String> _consumables = {};
  static Widget _popUpWidget;
  static AndroidValidation _androidValidation;
  static AccessCredentials _accessCredentials;

  // static IosValidation _iosValidation;

  static Future<void> init({
    List<String> productsIds = const [],
    List<String> subscriptionsIds = const [],
    Widget popUpWidget = const _DefaultPopUpWidget(),
    void Function(PurchasedItem purchasedItem) onPurchaseUpdated,
    AndroidValidation androidValidation,
    // IosValidation iosValidation,
  }) async {
    //! Set Private items
    _popUpWidget = popUpWidget;
    _onPurchaseUpdated = onPurchaseUpdated;
    _androidValidation = androidValidation;
    // _iosValidation = iosValidation;
    //! Init Connection
    await FlutterInappPurchase.instance.initConnection;
    await _refreshItemsAndroid();
    //! Create Listeners
    _purchaseErrorListen();
    _purchaseUpdateListen();
    _purchasePromotedListen();
    await _SimpleBillingData.init();
    //! Get IAPItems
    await _getProducts(productsIds);
    await _getSubscriptions(subscriptionsIds);
    //! Active Android Validation
    await _getClientCredentials();
    //! Get Offline Data
    getOfflinePurchases();
  }

  static Future getOfflinePurchases() async {
    Set<PurchasedItem> items;
    try {
      items =
          (await FlutterInappPurchase.instance.getAvailablePurchases()).toSet();
    } catch (e) {
      print('Failed getAvailablePurchases');
    }
    if (items != null) {
      await _SimpleBillingData.saveProducts({});
      for (final item in items) {
        final bool isSubscription = _subscriptionsItems.firstWhere(
                (e) => e.productId == item.productId,
                orElse: () => null) !=
            null;
        if (isSubscription) {
          _validateAndCallPurchaseUpdated(item);
        } else {
          purchases.add(item);
        }
      }
      await _SimpleBillingData.saveProducts(purchases);
    } else {
      purchases = _SimpleBillingData.products;
    }
  }

  //! Public Functions
  static Future<void> buyProduct(String productId,
      {@required BuildContext context, bool consumable = false}) async {
    if (consumable) {
      _consumables.add(productId);
    }
    _openLoading(context);
    try {
      await FlutterInappPurchase.instance.requestPurchase(productId);
    } catch (e) {
      _closeLoading();
    }
  }

  /// Request a subscription on `Android` or `iOS`.
  /// Result will be received
  /// in `purchaseUpdated` listener or `purchaseError` listener.
  ///
  /// **NOTICE** second parameter is required on `Android`.
  ///
  /// Identical to [requestPurchase] on `iOS`.
  static Future buySubscription(
    String productId, {
    @required BuildContext context,
    String oldSkuAndroid,
  }) async {
    _openLoading(context);
    try {
      await FlutterInappPurchase.instance
          .requestSubscription(productId, oldSkuAndroid: oldSkuAndroid);
    } catch (e) {
      _closeLoading();
    }
  }

  static Future<void> restorePurchases({@required BuildContext context}) async {
    _openLoading(context);
    try {
      final items = await FlutterInappPurchase.instance.getAvailablePurchases();
      items.forEach(_validateAndCallPurchaseUpdated);
    } finally {
      _closeLoading();
    }
  }

  static String getPrice(String productId) {
    return getProduct(productId)?.localizedPrice;
  }

  static IAPItem getProduct(String productId) {
    return _getProductById(productId) ?? _getSubscriptionById(productId);
  }

  static bool checkPurchase(String productId) {
    return purchases.firstWhere((e) => e.productId == productId,
            orElse: () => null) !=
        null;
  }

  static bool checkIfHaveAnyPurchase(List<String> productId) =>
      productId.map((e) => checkPurchase(e)).reduce((a, b) => a || b);

  static bool checkIfEverythingIsPurchased(List<String> productId) =>
      productId.map((e) => checkPurchase(e)).reduce((a, b) => a && b);

  //! Privates Functions

  static Future<bool> _validateReceiptAndroid(PurchasedItem purchased) async {
    if (_accessCredentials == null) return true;
    final bool isSubscription = _subscriptionsItems.firstWhere(
            (s) => s.productId == purchased.productId,
            orElse: () => null) !=
        null;
    final http.Response response = await validateReceiptAndroid(
      packageName: _androidValidation.packageName,
      accessToken: _accessCredentials.accessToken.data,
      productId: purchased.productId,
      productToken: purchased.purchaseToken,
      isSubscription: isSubscription,
    );
    if (response.reasonPhrase == 'OK') return true;
    return false;
  }

  /*static Future<bool> _validateReceiptIos(PurchasedItem purchased) async {
    if (iosValidation == null) return true;
    final receiptBody = {
      'receipt-data': purchased.transactionReceipt,
      'password': _iosValidation.password,
    };
    final Response response = await FlutterInappPurchase.instance
        .validateReceiptIos(receiptBody: receiptBody);
    print('validateReceiptIos response = $response');
    return response;
  }*/

  static void _closeLoading() {
    if (_context != null) {
      Navigator.pop(_context);
      _context = null;
    }
  }

  static void _openLoading(BuildContext context) {
    _context = context;
    if (_context != null) {
      showDialog(context: context, builder: (context) => _popUpWidget);
    }
  }

  static IAPItem _getProductById(String productId) => _productsItems
      .firstWhere((e) => e.productId == productId, orElse: () => null);

  static IAPItem _getSubscriptionById(String productId) => _subscriptionsItems
      .firstWhere((e) => e.productId == productId, orElse: () => null);

  static Future<void> _getSubscriptions(List<String> subscriptionsIds) async {
    if (subscriptionsIds.isNotEmpty) {
      try {
        _subscriptionsItems = await FlutterInappPurchase.instance
            .getSubscriptions(subscriptionsIds);
        _SimpleBillingData.saveIAPItems(_subscriptionsItems,
            isSubscription: true);
      } catch (e) {
        _subscriptionsItems =
            _SimpleBillingData.getIAPItems(isSubscription: true);
      }
    }
  }

  static Future<void> _getProducts(List<String> productsIds) async {
    if (productsIds.isNotEmpty) {
      try {
        _productsItems =
            await FlutterInappPurchase.instance.getProducts(productsIds);
        _SimpleBillingData.saveIAPItems(_productsItems);
      } catch (e) {
        _productsItems = _SimpleBillingData.getIAPItems();
      }
    }
  }

  static void _purchaseUpdateListen() {
    _purchaseUpdatedSubscription = FlutterInappPurchase.purchaseUpdated
        .listen((PurchasedItem purchase) async {
      _closeLoading();
      _validateAndCallPurchaseUpdated(purchase);
    });
  }

  static Future<void> _validateAndCallPurchaseUpdated(
      PurchasedItem purchase) async {
    bool isValidPurchase = Platform.isIOS;
    if (!isValidPurchase) {
      isValidPurchase = await _validateReceiptAndroid(purchase);
    }
    if (isValidPurchase) {
      purchases.add(purchase);
      final bool isConsumable = _consumables.contains(purchase.productId);
      if (isConsumable) {
        await FlutterInappPurchase.instance
            .consumePurchaseAndroid(purchase.purchaseToken);
      } else {
        _SimpleBillingData.saveProducts(purchases);
      }
      try {
        await FlutterInappPurchase.instance
            .finishTransaction(purchase, isConsumable: isConsumable);
      } catch (e) {}
      _onPurchaseUpdated(purchase);
    }
  }

  static void _purchaseErrorListen() {
    _purchaseErrorSubscription = FlutterInappPurchase.purchaseError
        .listen((PurchaseResult purchaseError) {
      _closeLoading();
    });
  }

  static void _purchasePromotedListen() {
    _purchasePromotedSubscription =
        FlutterInappPurchase.purchasePromoted.listen(
      (String onData) {},
    );
  }

  static Future<void> dispose() async {
    await FlutterInappPurchase.instance.endConnection;
    _purchaseUpdatedSubscription?.cancel();
    _purchaseErrorSubscription?.cancel();
    _purchasePromotedSubscription?.cancel();
  }

  static Future<void> _getClientCredentials() async {
    if (Platform.isAndroid && _androidValidation != null) {
      try {
        final scopes = ['https://www.googleapis.com/auth/androidpublisher'];
        final clienteCredentials =
            ServiceAccountCredentials.fromJson(_androidValidation.accessToken);
        final client = http.Client();
        _accessCredentials = await obtainAccessCredentialsViaServiceAccount(
            clienteCredentials, scopes, client);
      } catch (e) {}
    }
  }

  static Future<void> _refreshItemsAndroid() async {
    try {
      await FlutterInappPurchase.instance.consumeAllItems;
    } catch (err) {
      // ignore: avoid_print
      print('consumeAllItems error: $err');
    }
  }

  static Future<http.Response> validateReceiptAndroid({
    @required String packageName,
    @required String productId,
    @required String productToken,
    @required String accessToken,
    bool isSubscription = false,
  }) async {
    assert(packageName != null);
    assert(productId != null);
    assert(productToken != null);
    assert(accessToken != null);
    final String type = isSubscription ? 'subscriptions' : 'products';
    final String url =
        'https://www.googleapis.com/androidpublisher/v3/applications/$packageName/purchases/$type/$productId/tokens/$productToken?access_token=$accessToken';
    return http.get(
      url,
      headers: {
        'Accept': 'application/json',
      },
    );
  }
}

class AndroidValidation {
  AndroidValidation({@required this.packageName, @required this.accessToken});
  final String packageName;
  final Map<String, String> accessToken;
}

class IosValidation {
  IosValidation({@required this.password});
  final String password;
}
